"""."""

from mathphys.functions import get_namedtuple as _get_namedtuple

from ..namesys import SiriusPVName as _PVName
from .device import Device as _Device, DeviceSet as _DeviceSet


class Screen(_DeviceSet):
    """Screen Device."""

    class DEVICES:
        """Devices names."""

        TB_1 = 'TB-01:DI-Scrn-1'
        TB_2 = 'TB-01:DI-Scrn-2'
        TB_3 = 'TB-02:DI-Scrn-1'
        TB_4 = 'TB-02:DI-Scrn-2'
        TB_5 = 'TB-03:DI-Scrn'
        TB_6 = 'TB-04:DI-Scrn'
        BO_1 = 'BO-01D:DI-Scrn-1'
        BO_2 = 'BO-01D:DI-Scrn-2'
        BO_3 = 'BO-02U:DI-Scrn'
        TS_1 = 'TS-01:DI-Scrn'
        TS_2 = 'TS-02:DI-Scrn'
        TS_3 = 'TS-03:DI-Scrn'
        TS_4 = 'TS-04:DI-Scrn-1'
        TS_5 = 'TS-04:DI-Scrn-2'
        TS_6 = 'TS-04:DI-Scrn-3'
        LI_1 = 'LA-BI:PRF1'
        LI_2 = 'LA-BI:PRF2'
        LI_3 = 'LA-BI:PRF3'
        LI_4 = 'LA-BI:PRF4'
        LI_5 = 'LA-BI:PRF5'
        LI = (LI_1, LI_2, LI_3, LI_4, LI_5)
        TB = (TB_1, TB_2, TB_3, TB_4, TB_5, TB_6)
        BO = (BO_1, BO_2, BO_3)
        TS = (TS_1, TS_2, TS_3, TS_4, TS_5, TS_6)
        ALL = TB + BO + TS + LI

    def __new__(cls, devname):
        """."""
        # check if device exists
        if devname not in Screen.DEVICES.ALL:
            raise NotImplementedError(devname)

        if devname.lower().startswith('la'):
            return LIScreen(devname)
        else:
            return TBBOTSScreen(devname)


class LIScreen(_Device):
    """."""

    PROPERTIES_DEFAULT = (
        'RAW:ArrayData',
        'IMG:ArrayData',
        'ROI:MaxSizeX_RBV',
        'ROI:MaxSizeY_RBV',
        'ROI:SizeX',
        'ROI:SizeY',
        'ROI:SizeX_RBV',
        'ROI:SizeY_RBV',
        'ROI:MinX_',
        'ROI:MinY_',
        'ROI:MinX_RBV',
        'ROI:MinY_RBV',
        'X:Gauss:Peak',
        'Y:Gauss:Peak',
        'X:Gauss:Sigma',
        'Y:Gauss:Sigma',
        'X:Gauss:Coef',
        'Y:Gauss:Coef',
        'CAM:Gain',
        'CAM:Gain_RBV',
        'CAM:AcquireTime',
        'CAM:AcquireTime_RBV',
    )

    def __init__(self, devname, props2init='all'):
        """."""
        if devname not in Screen.DEVICES.ALL:
            raise NotImplementedError(devname)
        super().__init__(devname, props2init=props2init)

    @property
    def image_width(self):
        """."""
        return self['ROI:MaxSizeX_RBV']

    @property
    def image_height(self):
        """."""
        return self['ROI:MaxSizeY_RBV']

    @property
    def image(self):
        """."""
        row_image = self['RAW:ArrayData']
        h = self.image_height
        w = self.image_width
        return row_image.reshape([h, w])

    @property
    def centerx(self):
        """."""
        return self['X:Gauss:Peak']

    @property
    def centery(self):
        """."""
        return self['Y:Gauss:Peak']

    @property
    def sigmax(self):
        """."""
        return self['X:Gauss:Sigma']

    @property
    def sigmay(self):
        """."""
        return self['Y:Gauss:Sigma']

    @property
    def scale_factor_x(self):
        """Pixel to mm."""
        return self['X:Gauss:Coef']

    @property
    def scale_factor_y(self):
        """Pixel to mm."""
        return self['Y:Gauss:Coef']

    @property
    def roix_size(self):
        """."""
        return self['ROI:SizeX_RBV']

    @property
    def roiy_size(self):
        """."""
        return self['ROI:SizeY_RBV']

    @property
    def roix_start(self):
        """."""
        return self['ROI:MinX_RBV']

    @property
    def roiy_start(self):
        """."""
        return self['ROI:MinY_RBV']

    @property
    def image_roi(self):
        """."""
        img = self['IMG:ArrayData']
        return img.reshape((self.roix_size, self.roiy_size))

    @property
    def cam_exposure(self):
        """."""
        return self['CAM:AcquireTime_RBV']

    @property
    def cam_gain(self):
        """."""
        return self['CAM:Gain_RBV']


class TBBOTSScreen(_DeviceSet):
    """."""

    DsblEnbl = _get_namedtuple('DsblEnbl', ['OFF', 'ON'])
    ScrnPosition = _get_namedtuple(
        'ScrnPosition', ['None_', 'Calibration', 'Fluorescent', 'Unknown']
    )
    Homing = _get_namedtuple('Homing', ['In Progress', 'Finished', 'Error'])

    def __init__(self, devname):
        """."""
        # check if device exists
        if devname not in Screen.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        self.screen = _Screen(devname)
        self.screencam = _ScreenCam(devname)
        devs = [self.screen, self.screencam]
        super().__init__(devs, devname=devname)

    @property
    def homing_status_str(self):
        """."""
        return self.Homing._fields[self.screen['HomeDone-Mon']]

    @property
    def homing_status(self):
        """."""
        return self.screen['HomeDone-Mon']

    def cmd_do_homing(self, timeout=60):
        """."""
        self.screen['Home-Cmd'] = 1
        return self.screen.wait(
            'HomeDone-Mon', self.Homing.Finished, timeout=timeout
        )

    @property
    def screen_position_str(self):
        """."""
        return self.ScrnPosition._fields[self.screen['ScrnType-Sts']]

    @property
    def screen_position(self):
        """."""
        return self.screen['ScrnType-Sts']

    @screen_position.setter
    def screen_position(self, value):
        """."""
        self.screen._enum_setter('ScrnType-Sel', value, self.ScrnPosition)

    def change_screen_position(self, screen_type, timeout=20):
        """."""
        screen_type = self.screen._enum_selector(
            screen_type, self.ScrnPosition
        )
        self.screen_position = screen_type
        return self.screen.wait('ScrnType-Sts', screen_type, timeout=timeout)

    def move_to_calibration(self, timeout=20):
        """."""
        return self.change_screen_position(
            self.ScrnPosition.Calibration, timeout=timeout
        )

    def move_to_fluorescent(self, timeout=20):
        """."""
        return self.change_screen_position(
            self.ScrnPosition.Fluorescent, timeout=timeout
        )

    def move_to_none(self, timeout=20):
        """."""
        return self.change_screen_position(
            self.ScrnPosition.None_, timeout=timeout
        )

    @property
    def enabled_str(self):
        """."""
        return self.DsblEnbl._fields[self.screencam['Enbl-Sts']]

    @property
    def enabled(self):
        """."""
        return self.screencam['Enbl-Sts']

    @enabled.setter
    def enabled(self, value):
        """."""
        self.screencam._enum_setter('Enbl-Sel', value, self.DsblEnbl)

    def cmd_enable(self, timeout=1):
        """."""
        self.enabled = self.DsblEnbl.ON
        return self.screencam.wait(
            'Enbl-Sts', self.DsblEnbl.ON, timeout=timeout
        )

    def cmd_disable(self, timeout=1):
        """."""
        self.enabled = self.DsblEnbl.OFF
        return self.screencam.wait(
            'Enbl-Sts', self.DsblEnbl.OFF, timeout=timeout
        )

    @property
    def image_width(self):
        """."""
        return self.screen['ImgROIWidth-RB']

    @property
    def image_height(self):
        """."""
        return self.screen['ImgROIHeight-RB']

    @property
    def image(self):
        """."""
        row_image = self.screen['ImgData-Mon']
        h = self.image_height
        w = self.image_width
        return row_image.reshape([h, w])

    @property
    def centerx(self):
        """."""
        return self.screen['CenterXDimFei-Mon']

    @property
    def centery(self):
        """."""
        return self.screen['CenterYDimFei-Mon']

    @property
    def sigmax(self):
        """."""
        return self.screen['SigmaXDimFei-Mon']

    @property
    def sigmay(self):
        """."""
        return self.screen['SigmaYDimFei-Mon']

    @property
    def angle(self):
        """."""
        return self.screen['ThetaDimFei-Mon']

    @property
    def exposure_time(self):
        """."""
        return self.screencam['ExposureTime-RB']

    @exposure_time.setter
    def exposure_time(self, value):
        """."""
        self.screencam['ExposureTime-SP'] = value

    @property
    def gain(self):
        """."""
        return self.screencam['Gain-RB']

    @gain.setter
    def gain(self, value):
        """."""
        self.screencam['Gain-SP'] = value

    @property
    def frame_count(self):
        """."""
        return self.screencam['FrameCnt-Mon']

    @property
    def scale_factor_x(self):
        """Pixel to mm."""
        return self.screencam['ScaleFactorX-RB']

    @property
    def scale_factor_y(self):
        """Pixel to mm."""
        return self.screencam['ScaleFactorY-RB']

    @property
    def center_offset_x(self):
        """."""
        return self.screencam['CenterOffsetX-RB']

    @property
    def center_offset_y(self):
        """."""
        return self.screencam['CenterOffsetY-RB']


""" TODO: _Screen and _ScreenCam class could be merged into DVF class, since
    they seem to implement the same functionalities and originally had the same
    base IOC code.
"""


class _Screen(_Device):
    """."""

    DEVICES = Screen.DEVICES

    PROPERTIES_DEFAULT = (
        'Home-Cmd',
        'HomeDone-Mon',
        'ScrnType-Sel',
        'ScrnType-Sts',
        'ImgData-Mon',
        'CenterXDimFei-Mon',
        'CenterYDimFei-Mon',
        'SigmaXDimFei-Mon',
        'SigmaYDimFei-Mon',
        'ThetaDimFei-Mon',
        'ImgROIHeight-RB',
        'ImgROIWidth-RB',
    )

    def __init__(self, devname, props2init='all'):
        """."""
        if devname not in Screen.DEVICES.ALL:
            raise NotImplementedError(devname)
        super().__init__(devname, props2init=props2init)


class _ScreenCam(_Device):
    """."""

    DEVICES = Screen.DEVICES

    PROPERTIES_DEFAULT = (
        'Enbl-Sel',
        'Enbl-Sts',
        'ScaleFactorX-RB',
        'ScaleFactorY-RB',
        'CenterOffsetX-RB',
        'CenterOffsetY-RB',
        'ExposureTime-SP',
        'ExposureTime-RB',
        'Gain-SP',
        'Gain-RB',
        'FrameCnt-Mon',
    )

    def __init__(self, devname, props2init='all'):
        """."""
        if devname not in Screen.DEVICES.ALL:
            raise NotImplementedError(devname)
        devname = _PVName(devname).substitute(dev='ScrnCam')
        super().__init__(devname, props2init=props2init)



"""Devices subpackage."""

from .afc_acq_core import AFCACQLogicalTrigger, AFCPhysicalTrigger
from .bbb import BunchbyBunch
from .beamline import CAXCtrl
from .blm import BLM, FamBLMs
from .bpm import BPM
from .bpm_eq import EqualizeBPMs
from .bpm_fam import FamBPMs
from .currinfo import (
    CurrInfoAS,
    CurrInfoBO,
    CurrInfoLinear,
    CurrInfoSI,
    CurrInfoTranspEff
)
from .dcct import DCCT
from .device import Device, DeviceSet
from .dvf import DVF, DVFImgProc
from .egun import EGBias, EGFilament, EGHVPS, EGPulsePS, EGTriggerPS, EGun
from .energy import Energy
from .fofb import (
    BPMDCC,
    FamFastCorrs,
    FamFOFBControllers,
    FOFBCtrlDCC,
    FOFBCtrlRef,
    HLFOFB
)
from .fofb_acq import (
    FamFOFBLamp,
    FamFOFBSysId,
    FOFBCtrlLamp,
    FOFBCtrlSysId,
    FOFBPSLamp,
    FOFBPSSysId
)
from .fpmosc import FPMOsc
from .gamma_monitor import FamGammaMonitors, GammaCounter, GammaMonitor
from .ict import ICT, TranspEff
from .idff import IDFF, IDFFCtrl, IDFFCtrlHard, IDFFCtrlSoft
from .ids import APU, DELTA, EPU, ID, IDBase, IVU, PAPU, VPU, WIG
from .injctrl import InjCtrl
from .injsys import (
    BOPSRampStandbyHandler,
    BORFRampStandbyHandler,
    InjSysPUModeHandler,
    InjSysStandbyHandler,
    LinacStandbyHandler,
    PUMagsStandbyHandler
)
from .intlkctrl import ASMPSCtrl, ASPPSCtrl, BLInterlockCtrl
from .lienergy import LIEnergy
from .lillrf import DevLILLRF, LILLRF
from .machshift import MachShift
from .modltr import LIModltr
from .orbit_interlock import (
    BaseOrbitIntlk,
    BPMOrbitIntlk,
    HLOrbitInterlock,
    OrbitInterlock
)
from .posang import PosAng
from .psconv import PSProperty, StrengthConv
from .pssofb import PSApplySOFB, PSCorrSOFB
from .pwrsupply import (
    PowerSupply,
    PowerSupplyFBP,
    PowerSupplyFC,
    PowerSupplyPU
)
from .rf import (
    ASLLRF,
    BOLLRFPreAmp,
    BORF300VDCAmp,
    BORFCavMonitor,
    BORFDCAmp,
    RFCav,
    RFGen,
    RFKillBeam,
    SILLRFPreAmp,
    SIRFACAmp,
    SIRFCavMonitor,
    SIRFDCAmp
)
from .scraper import ScraperH, ScraperV
from .screen import Screen
from .sofb import SOFB
from .syncd import DevicesSync
from .timing import Event, EVG, HLTiming, Trigger
from .tune import Tune, TuneCorr, TuneFrac, TuneProc

del device, bpm, dcct, egun, ict, lillrf, modltr
del pwrsupply, posang, psconv, pssofb, rf, injsys, injctrl
del screen, tune, sofb, timing, syncd, energy
del ids, currinfo, bbb, machshift, dvf, lienergy, intlkctrl, scraper
del blm, beamline



"""Test."""

try:
    from matplotlib import pyplot as plt
except:
    plt = None

import numpy as np
from siriuspy.clientarch.time import Time
from siriuspy.machshift.macreport import MacReport

# failure statistics per month
intervals = [
    # [Time(2023, 1, 1, 0, 0), Time(2023, 1, 31, 23, 59, 59)],
    # [Time(2023, 2, 1, 0, 0), Time(2023, 2, 28, 23, 59, 59)],
    # [Time(2023, 3, 1, 0, 0), Time(2023, 3, 31, 23, 59, 59)],
    # [Time(2023, 4, 1, 0, 0), Time(2023, 4, 30, 23, 59, 59)],
    # [Time(2023, 5, 1, 0, 0), Time(2023, 5, 31, 23, 59, 59)],
    # [Time(2023, 6, 1, 0, 0), Time(2023, 6, 30, 23, 59, 59)],
    # [Time(2023, 7, 1, 0, 0), Time(2023, 7, 31, 23, 59, 59)],
    # [Time(2023, 8, 1, 0, 0), Time(2023, 8, 31, 23, 59, 59)],
    # [Time(2023, 9, 1, 0, 0), Time(2023, 9, 30, 23, 59, 59)],
    # [Time(2023, 10, 1, 0, 0), Time(2023, 10, 31, 23, 59, 59)],
    # [Time(2023, 11, 1, 0, 0), Time(2023, 11, 30, 23, 59, 59)],
    # [Time(2023, 12, 1, 0, 0), Time(2023, 12, 31, 23, 59, 59)],

    # [Time(2024, 1, 1, 0, 0), Time(2024, 1, 31, 23, 59, 59)],
    # [Time(2024, 2, 1, 0, 0), Time(2024, 2, 29, 23, 59, 59)],
    # [Time(2024, 3, 1, 0, 0), Time(2024, 3, 31, 23, 59, 59)],
    # [Time(2024, 4, 1, 0, 0), Time(2024, 4, 30, 23, 59, 59)],
    # [Time(2024, 5, 1, 0, 0), Time(2024, 5, 31, 23, 59, 59)],
    # [Time(2024, 6, 1, 0, 0), Time(2024, 6, 30, 23, 59, 59)],
    # [Time(2024, 7, 1, 0, 0), Time(2024, 7, 31, 23, 59, 59)],
    # [Time(2024, 8, 1, 0, 0), Time(2024, 8, 31, 23, 59, 59)],
    # [Time(2024, 9, 1, 0, 0), Time(2024, 9, 30, 23, 59, 59)],
    # [Time(2024, 10, 1, 0, 0), Time(2024, 10, 31, 23, 59, 59)],
    # [Time(2024, 11, 1, 0, 0), Time(2024, 11, 30, 23, 59, 59)],
    # [Time(2024, 12, 1, 0, 0), Time(2024, 12, 31, 23, 59, 59)],

    [Time(2025, 1, 1, 0, 0), Time(2025, 1, 31, 23, 59, 59)],
    [Time(2025, 2, 1, 0, 0), Time(2025, 2, 28, 23, 59, 59)],
    [Time(2025, 3, 1, 0, 0), Time(2025, 3, 31, 23, 59, 59)],
    [Time(2025, 4, 1, 0, 0), Time(2025, 4, 30, 23, 59, 59)],
    [Time(2025, 5, 1, 0, 0), Time(2025, 5, 31, 23, 59, 59)],
    [Time(2025, 6, 1, 0, 0), Time(2025, 6, 30, 23, 59, 59)],
    [Time(2025, 7, 1, 0, 0), Time(2025, 7, 31, 23, 59, 59)],
    [Time(2025, 8, 1, 0, 0), Time(2025, 8, 31, 23, 59, 59)],
    [Time(2025, 9, 1, 0, 0), Time(2025, 9, 30, 23, 59, 59)],
    [Time(2025, 10, 1, 0, 0), Time(2025, 10, 31, 23, 59, 59)],
]

macreports = dict()
for intvl in intervals:
    macreports[intvl[0]] = MacReport()
    # macreports[intvl[0]].connector.server_url = 'https://archiver-temp.cnpem.br'   # necessary for 2024 and early
    macreports[intvl[0]].connector.timeout = 30
    macreports[intvl[0]].time_start = intvl[0]
    macreports[intvl[0]].time_stop = intvl[1]
    macreports[intvl[0]].update()

mtbfs, mttrs, reliabs = dict(), dict(), dict()
progrmd, delivd, usertot = dict(), dict(), dict()
currmean, currstd = dict(), dict()
stable, unstable, relstable = dict(), dict(), dict()
nrbeamdump = dict()
for date, macr in macreports.items():
    mtbfs[date] = macr.usershift_time_between_failures_average
    mttrs[date] = macr.usershift_time_to_recover_average
    reliabs[date] = macr.usershift_beam_reliability
    progrmd[date] = macr.usershift_progmd_time
    delivd[date] = macr.usershift_delivd_time
    usertot[date] = macr.usershift_total_time
    currmean[date] = macr.usershift_current_average
    currstd[date] = macr.usershift_current_stddev
    stable[date] = macr.usershift_total_stable_beam_time
    unstable[date] = macr.usershift_total_unstable_beam_time
    relstable[date] = macr.usershift_relative_stable_beam_time
    nrbeamdump[date] = macr.usershift_beam_dump_count

# create table to help visualization
str_ = '{:<10s}' + '{:>16s}'*10 + '{:>20s}'
print(str_.format(
    'Y-M', 'MTBF', 'MTTR',
    'Reliability', 'Progrmd hours', 'Delivd hours', 'Total hours',
    '% stable hours', 'Stable hours', 'Unstable hours', '# Beam Dump',
    'Current (Avg±Std)'))
str_ = '{:<10s}' + '    {:>12.3f}'*10 + '    {:4.3f} ± {:4.3f}'
for date in macreports:
    print(str_.format(
        str(date.year)+'-'+str(date.month),
        mtbfs[date],
        mttrs[date],
        reliabs[date],
        progrmd[date],
        delivd[date],
        usertot[date],
        relstable[date],
        stable[date],
        unstable[date],
        nrbeamdump[date],
        currmean[date], currstd[date],
    ))

fig, axs = plt.subplots(3, 1, sharex=True)
fig.set_size_inches(9, 6)
fig.subplots_adjust(top=0.96, left=0.08, bottom=0.05, right=0.96)
axs[0].xaxis.axis_date()
axs[0].plot(mtbfs.keys(), mtbfs.values(), '-b')
axs[0].plot(mtbfs.keys(), mtbfs.values(), 'ob')
axs[0].set_title('MTBF')
axs[0].grid()
axs[1].xaxis.axis_date()
axs[1].plot(mttrs.keys(), mttrs.values(), '-r')
axs[1].plot(mttrs.keys(), mttrs.values(), 'or')
axs[1].set_title('MTTR')
axs[1].grid()
axs[2].xaxis.axis_date()
axs[2].plot(reliabs.keys(), reliabs.values(), '-g')
axs[2].plot(reliabs.keys(), reliabs.values(), 'og')
axs[2].set_title('Reliability')
axs[2].grid()
fig.show()

# plots to help diagnostics
report = macreports[Time(2025, 5, 1, 0, 0)]

# - verify beam dump
datetimes = np.array([Time(t) for t in report._raw_data['Timestamp']])
beam_dump_values = report._beam_dump_values
plt.plot(datetimes, beam_dump_values)
plt.show()

# - plot any month data
report.plot_raw_data().show()


# programmed vs. delivered hours
macr = MacReport()
macr.connector.timeout = 300
macr.time_start = Time(2024, 1, 1, 0, 0)
macr.time_stop = Time(2024, 12, 31, 23, 59, 59)
macr.update()

str_ = '{:<10s}' + '{:>16s}'*10 + '{:>20s}'
print(str_.format(
    'Y-M', 'MTBF', 'MTTR',
    'Reliability', 'Progrmd hours', 'Delivd hours', 'Total hours',
    '% stable hours', 'Stable hours', 'Unstable hours', '# Beam Dump',
    'Current (Avg±Std)'))
str_ = '{:<10s}' + '    {:>12.3f}'*10 + '    {:4.3f} ± {:4.3f}'
print(
    str_.format(
        str(macr.time_start.year)+'-'+str(macr.time_start.month) + ' a ' +
        str(macr.time_stop.year)+'-'+str(macr.time_stop.month),
        macr.usershift_time_between_failures_average,
        macr.usershift_time_to_recover_average,
        macr.usershift_beam_reliability,
        macr.usershift_progmd_time,
        macr.usershift_delivd_time,
        macr.usershift_total_time,
        macr.usershift_relative_stable_beam_time,
        macr.usershift_total_stable_beam_time,
        macr.usershift_total_unstable_beam_time,
        macr.usershift_beam_dump_count,
        macr.usershift_current_average,
        macr.usershift_current_stddev,
    ))

rd = macr.raw_data
dtimes = np.diff(rd['Timestamp'])
dtimes = np.r_[dtimes, dtimes[-1]]
dtimes = dtimes/60/60

dates = [Time(d) for d in rd['Timestamp']]

dtimes_users_progmd = dtimes*rd['UserShiftProgmd']
cum_progmd = np.cumsum(dtimes_users_progmd)
dtimes_users_impltd = dtimes*rd['UserShiftDelivd']
cum_deliv = np.cumsum(dtimes_users_impltd)

fig = plt.figure()
axs = plt.gca()
axs.xaxis.axis_date()
axs.plot(dates, cum_progmd, '-', label='Programmed')
axs.plot(dates, cum_deliv, '-', label='Delivered')
axs.grid()
axs.set_ylabel('Integrated Hours')
plt.legend(loc=4)
plt.title('Integrated User Hours')
fig.show()



"""."""

import time as _time

import numpy as _np

from ..sofb.csdev import SOFBFactory
from ..sofb.utils import si_calculate_bump as _si_calculate_bump
from .device import Device as _Device


class SOFB(_Device):
    """SOFB Device."""

    class DEVICES:
        """Devices names."""

        TB = 'TB-Glob:AP-SOFB'
        BO = 'BO-Glob:AP-SOFB'
        TS = 'TS-Glob:AP-SOFB'
        SI = 'SI-Glob:AP-SOFB'
        ALL = (TB, BO, TS, SI)

    def __new__(
        cls,
        devname,
        props2init='all',
        auto_monitor=True,
        auto_monitor_mon=False,
    ):
        """."""
        # check if device exists
        if devname not in SOFB.DEVICES.ALL:
            raise NotImplementedError(devname)

        # SOFB object
        data = SOFBFactory.create(devname[:2])

        if data.acc == 'SI':
            return SISOFB(
                devname,
                data,
                props2init=props2init,
                auto_monitor=auto_monitor,
                auto_monitor_mon=auto_monitor_mon,
            )
        elif data.isring:
            return BOSOFB(
                devname,
                data,
                props2init=props2init,
                auto_monitor=auto_monitor,
                auto_monitor_mon=auto_monitor_mon,
            )
        return TLSOFB(
            devname,
            data,
            props2init=props2init,
            auto_monitor=auto_monitor,
            auto_monitor_mon=auto_monitor_mon,
        )


class TLSOFB(_Device):
    """SOFB Device."""

    PROPERTIES_DEFAULT = (
        'TrigAcqChan-Sel',
        'TrigAcqChan-Sts',
        'OrbStatus-Mon',
        'RespMat-SP',
        'RespMat-RB',
        'RespMat-Mon',
        'InvRespMat-Mon',
        'KickCH-Mon',
        'KickCV-Mon',
        'DeltaKickCH-Mon',
        'DeltaKickCV-Mon',
        'DeltaKickCH-SP',
        'DeltaKickCV-SP',
        'MaxDeltaKickCH-RB',
        'MaxDeltaKickCV-RB',
        'MaxDeltaKickCH-SP',
        'MaxDeltaKickCV-SP',
        'ManCorrGainCH-SP',
        'ManCorrGainCV-SP',
        'ManCorrGainCH-RB',
        'ManCorrGainCV-RB',
        'RefOrbX-SP',
        'RefOrbY-SP',
        'RefOrbX-RB',
        'RefOrbY-RB',
        'BPMXEnblList-SP',
        'BPMYEnblList-SP',
        'BPMXEnblList-RB',
        'BPMYEnblList-RB',
        'CHEnblList-SP',
        'CVEnblList-SP',
        'CHEnblList-RB',
        'CVEnblList-RB',
        'CalcDelta-Cmd',
        'ApplyDelta-Cmd',
        'SmoothReset-Cmd',
        'ApplyDelta-Mon',
        'SmoothNrPts-SP',
        'SmoothNrPts-RB',
        'BufferCount-Mon',
        'TrigNrSamplesPre-SP',
        'TrigNrSamplesPre-RB',
        'TrigNrSamplesPost-SP',
        'TrigNrSamplesPost-RB',
        'SPassSum-Mon',
        'SPassOrbX-Mon',
        'SPassOrbY-Mon',
        'MeasRespMat-Cmd',
        'MeasRespMat-Mon',
        'MeasRespMatKickCH-SP',
        'MeasRespMatKickCH-RB',
        'MeasRespMatKickCV-SP',
        'MeasRespMatKickCV-RB',
        'MeasRespMatWait-SP',
        'MeasRespMatWait-RB',
        'NrSingValues-Mon',
        'MinSingValue-SP',
        'MinSingValue-RB',
        'TrigAcqCtrl-Sel',
        'TrigAcqCtrl-Sts',
        'TrigAcqConfig-Cmd',
        'SyncBPMs-Cmd',
    )

    _default_timeout = 10  # [s]
    _default_timeout_respm = 2 * 60 * 60  # [s]
    _default_timeout_kick_apply = 2  # [s]

    def __init__(
        self,
        devname,
        data=None,
        props2init='all',
        auto_monitor=True,
        auto_monitor_mon=False,
    ):
        """."""
        # check if device exists
        if devname not in SOFB.DEVICES.ALL:
            raise NotImplementedError(devname)

        # SOFB object
        self._data = data or SOFBFactory.create(devname[:2])
        # call base class constructor
        super().__init__(
            devname,
            props2init=props2init,
            auto_monitor=auto_monitor,
            auto_monitor_mon=auto_monitor_mon,
        )

    @property
    def data(self):
        """."""
        return self._data

    @property
    def trigchannel(self):
        """."""
        return self['TrigAcqChan-Sts']

    @trigchannel.setter
    def trigchannel(self, value):
        self._enum_setter('TrigAcqChan-Sel', value, self._data.TrigAcqChan)

    @property
    def respmat(self):
        """Raw response matrix."""
        return self['RespMat-RB'].reshape(self._data.nr_bpms * 2, -1)

    @respmat.setter
    def respmat(self, mat):
        self['RespMat-SP'] = _np.array(mat).ravel()

    @property
    def respmat_mon(self):
        """Applied response matrix."""
        return self['RespMat-Mon'].reshape(self._data.nr_bpms * 2, -1)

    @property
    def invrespmat(self):
        """Inverse response matrix."""
        return self['InvRespMat-Mon'].reshape(-1, self._data.nr_bpms * 2)

    @property
    def trigchannel_str(self):
        """."""
        return self._data.TrigAcqChan._fields[self['TrigAcqChan-Sts']]

    @property
    def sp_trajx(self):
        """."""
        return self['SPassOrbX-Mon']

    @property
    def sp_trajy(self):
        """."""
        return self['SPassOrbY-Mon']

    @property
    def sp_sum(self):
        """."""
        return self['SPassSum-Mon']

    @property
    def trajx(self):
        """."""
        return self.sp_trajx

    @property
    def trajy(self):
        """."""
        return self.sp_trajy

    @property
    def sum(self):
        """."""
        return self.sp_sum

    @property
    def kickch(self):
        """."""
        return self['KickCH-Mon']

    @property
    def kickcv(self):
        """."""
        return self['KickCV-Mon']

    @property
    def deltakickch(self):
        """."""
        return self['DeltaKickCH-Mon']

    @deltakickch.setter
    def deltakickch(self, value):
        """."""
        self['DeltaKickCH-SP'] = value

    @property
    def deltakickcv(self):
        """."""
        return self['DeltaKickCV-Mon']

    @deltakickcv.setter
    def deltakickcv(self, value):
        """."""
        self['DeltaKickCV-SP'] = value

    @property
    def maxdeltakickch(self):
        """."""
        return self['MaxDeltaKickCH-RB']

    @maxdeltakickch.setter
    def maxdeltakickch(self, value):
        """."""
        self['MaxDeltaKickCH-SP'] = value

    @property
    def maxdeltakickcv(self):
        """."""
        return self['MaxDeltaKickCV-RB']

    @maxdeltakickcv.setter
    def maxdeltakickcv(self, value):
        """."""
        self['MaxDeltaKickCV-SP'] = value

    @property
    def mancorrgainch(self):
        """."""
        return self['ManCorrGainCH-RB']

    @mancorrgainch.setter
    def mancorrgainch(self, value):
        """."""
        self['ManCorrGainCH-SP'] = value

    @property
    def mancorrgaincv(self):
        """."""
        return self['ManCorrGainCV-RB']

    @mancorrgaincv.setter
    def mancorrgaincv(self, value):
        """."""
        self['ManCorrGainCV-SP'] = value

    @property
    def refx(self):
        """."""
        return self['RefOrbX-RB']

    @refx.setter
    def refx(self, value):
        """."""
        self['RefOrbX-SP'] = value

    @property
    def refy(self):
        """."""
        return self['RefOrbY-RB']

    @refy.setter
    def refy(self, value):
        """."""
        self['RefOrbY-SP'] = value

    @property
    def bpmxenbl(self):
        """."""
        return _np.array(self['BPMXEnblList-RB'], dtype=bool)

    @bpmxenbl.setter
    def bpmxenbl(self, value):
        """."""
        self['BPMXEnblList-SP'] = value

    @property
    def bpmyenbl(self):
        """."""
        return _np.array(self['BPMYEnblList-RB'], dtype=bool)

    @bpmyenbl.setter
    def bpmyenbl(self, value):
        """."""
        self['BPMYEnblList-SP'] = value

    @property
    def chenbl(self):
        """."""
        return _np.array(self['CHEnblList-RB'], dtype=bool)

    @chenbl.setter
    def chenbl(self, value):
        """."""
        self['CHEnblList-SP'] = value

    @property
    def cvenbl(self):
        """."""
        return _np.array(self['CVEnblList-RB'], dtype=bool)

    @cvenbl.setter
    def cvenbl(self, value):
        """."""
        self['CVEnblList-SP'] = value

    @property
    def buffer_count(self):
        """."""
        return self['BufferCount-Mon']

    @property
    def nr_points(self):
        """."""
        return self['SmoothNrPts-RB']

    @nr_points.setter
    def nr_points(self, value):
        """."""
        self['SmoothNrPts-SP'] = int(value)

    @property
    def nr_singvals(self):
        """."""
        return self['NrSingValues-Mon']

    @property
    def singval_min(self):
        """."""
        return self['MinSingValue-RB']

    @singval_min.setter
    def singval_min(self, value):
        """."""
        self['MinSingValue-SP'] = value

    @property
    def trigacq(self):
        """."""
        return self['TrigAcqCtrl-Sts']

    @property
    def trigsamplepre(self):
        """."""
        return self['TrigNrSamplesPre-RB']

    @trigsamplepre.setter
    def trigsamplepre(self, value):
        """."""
        self['TrigNrSamplesPre-SP'] = int(value)

    @property
    def trigsamplepost(self):
        """."""
        return self['TrigNrSamplesPost-RB']

    @trigsamplepost.setter
    def trigsamplepost(self, value):
        """."""
        self['TrigNrSamplesPost-SP'] = int(value)

    def cmd_reset(self):
        """."""
        prop = 'SmoothReset-Cmd'
        val = self[prop]
        self[prop] = 1
        return self.wait(prop, val + 1)

    def cmd_calccorr(self):
        """."""
        prop = 'CalcDelta-Cmd'
        val = self[prop]
        self[prop] = 1
        return self.wait(prop, val + 1)

    def cmd_applycorr_ch(self):
        """."""
        prop = 'ApplyDelta-Cmd'
        val = self[prop]
        self[prop] = self._data.ApplyDelta.CH
        return self.wait(prop, val + 1)

    def cmd_applycorr_cv(self):
        """."""
        prop = 'ApplyDelta-Cmd'
        val = self[prop]
        self[prop] = self._data.ApplyDelta.CV
        return self.wait(prop, val + 1)

    def cmd_applycorr_rf(self):
        """."""
        prop = 'ApplyDelta-Cmd'
        val = self[prop]
        self[prop] = self._data.ApplyDelta.RF
        return self.wait(prop, val + 1)

    def cmd_applycorr_all(self):
        """."""
        prop = 'ApplyDelta-Cmd'
        val = self[prop]
        self[prop] = self._data.ApplyDelta.All
        return self.wait(prop, val + 1)

    def cmd_measrespmat_start(self):
        """."""
        prop = 'MeasRespMat-Cmd'
        val = self[prop]
        self[prop] = 0
        return self.wait(prop, val + 1)

    def cmd_measrespmat_stop(self):
        """."""
        prop = 'MeasRespMat-Cmd'
        val = self[prop]
        self[prop] = 1
        return self.wait(prop, val + 1)

    def cmd_measrespmat_reset(self):
        """."""
        prop = 'MeasRespMat-Cmd'
        val = self[prop]
        self[prop] = 2
        return self.wait(prop, val + 1)

    def cmd_trigacq_start(self, timeout=10):
        """."""
        self['TrigAcqCtrl-Sel'] = 'Start'
        ret = self.wait(
            'TrigAcqCtrl-Sts', self._data.TrigAcqCtrl.Start, timeout=timeout
        )
        if not ret:
            return False
        _time.sleep(0.6)  # Status PV updates at 2Hz
        return self.wait_orb_status_ok(timeout=timeout)

    def cmd_trigacq_stop(self, timeout=10):
        """."""
        self['TrigAcqCtrl-Sel'] = 'Stop'
        ret = self.wait(
            'TrigAcqCtrl-Sts', self._data.TrigAcqCtrl.Stop, timeout=timeout
        )
        if not ret:
            return False
        _time.sleep(0.6)  # Status PV updates at 2Hz
        return self.wait_orb_status_ok(timeout=timeout)

    def cmd_trigacq_config(self, timeout=10):
        """."""
        self['TrigAcqConfig-Cmd'] = 1
        return self.wait_orb_status_ok(timeout=timeout)

    @property
    def applydeltakick_mon(self):
        """."""
        return self['ApplyDelta-Mon']

    @property
    def measrespmat_mon(self):
        """."""
        return self['MeasRespMat-Mon']

    @property
    def measrespmat_kickch(self):
        """."""
        return self['MeasRespMatKickCH-RB']

    @measrespmat_kickch.setter
    def measrespmat_kickch(self, value):
        self['MeasRespMatKickCH-SP'] = value

    @property
    def measrespmat_kickcv(self):
        """."""
        return self['MeasRespMatKickCV-RB']

    @measrespmat_kickcv.setter
    def measrespmat_kickcv(self, value):
        self['MeasRespMatKickCV-SP'] = value

    @property
    def measrespmat_wait(self):
        """."""
        return self['MeasRespMatWait-RB']

    @measrespmat_wait.setter
    def measrespmat_wait(self, value):
        self['MeasRespMatWait-SP'] = value

    def correct_orbit_manually(self, nr_iters=10, residue=5):
        """."""
        self.cmd_turn_off_autocorr()
        for i in range(nr_iters):
            resx = self.orbx - self.refx
            resy = self.orby - self.refy
            resx = resx[self.bpmxenbl.nonzero()[0]]
            resy = resy[self.bpmyenbl.nonzero()[0]]
            resx = _np.sqrt(_np.sum(resx * resx) / resx.size)
            resy = _np.sqrt(_np.sum(resy * resy) / resy.size)
            if resx < residue and resy < residue:
                break
            self.wait_buffer()
            self.cmd_calccorr()
            self.cmd_applycorr_all()
            self.wait_apply_delta_kick()
            self.cmd_reset()
        return i, resx, resy

    def wait_buffer(self, timeout=None):
        """."""
        timeout = timeout or self._default_timeout
        return self.wait(
            'BufferCount-Mon', self.nr_points, timeout=timeout, comp='ge'
        )

    def wait_apply_delta_kick(self, timeout=None):
        """."""
        def_timeout = min(1.05 * self.deltakickrf, self.maxdeltakickrf) // 20
        def_timeout = max(self._default_timeout_kick_apply, def_timeout)
        timeout = timeout or def_timeout
        return self.wait(
            'ApplyDelta-Mon',
            self._data.ApplyDeltaMon.Applying,
            timeout=timeout,
            comp='ne',
        )

    def wait_respm_meas(self, timeout=None):
        """."""
        timeout = timeout or self._default_timeout_respm
        return self.wait(
            'MeasRespMat-Mon',
            self._data.MeasRespMatMon.Measuring,
            timeout=timeout,
            comp='ne',
        )

    def wait_orb_status_ok(self, timeout=10):
        """."""
        return self.wait('OrbStatus-Mon', 0, timeout=timeout)

    def cmd_sync_bpms(self):
        """Synchronize BPMs."""
        prop = 'SyncBPMs-Cmd'
        val = self[prop]
        self[prop] = 1
        return self.wait(prop, val + 1)


class BOSOFB(TLSOFB):
    """SOFB Device."""

    PROPERTIES_DEFAULT = TLSOFB.PROPERTIES_DEFAULT + (
        'MTurnAcquire-Cmd',
        'MTurnSum-Mon',
        'MTurnOrbX-Mon',
        'MTurnOrbY-Mon',
        'MTurnIdxOrbX-Mon',
        'MTurnIdxOrbY-Mon',
        'MTurnIdxSum-Mon',
        'MTurnTime-Mon',
        'MTurnDownSample-SP',
        'MTurnDownSample-RB',
        'MTurnIdx-SP',
        'MTurnIdx-RB',
        'MTurnIdxTime-Mon',
        'MTurnUseMask-Sel',
        'MTurnUseMask-Sts',
        'MTurnMaskSplBeg-SP',
        'MTurnMaskSplBeg-RB',
        'MTurnMaskSplEnd-SP',
        'MTurnMaskSplEnd-RB',
    )

    @property
    def mt_trajx(self):
        """."""
        return self['MTurnOrbX-Mon']

    @property
    def mt_trajy(self):
        """."""
        return self['MTurnOrbY-Mon']

    @property
    def mt_sum(self):
        """."""
        return self['MTurnSum-Mon']

    @property
    def mt_time(self):
        """."""
        return self['MTurnTime-Mon']

    @property
    def mt_trajx_idx(self):
        """."""
        return self['MTurnIdxOrbX-Mon']

    @property
    def mt_trajy_idx(self):
        """."""
        return self['MTurnIdxOrbY-Mon']

    @property
    def mt_sum_idx(self):
        """."""
        return self['MTurnIdxSum-Mon']

    @property
    def mt_downsampling(self):
        """."""
        return self['MTurnDownSample-RB']

    @mt_downsampling.setter
    def mt_downsampling(self, value):
        """."""
        self['MTurnDownSample-SP'] = int(value)

    @property
    def mt_idx(self):
        """."""
        return self['MTurnIdx-RB']

    @mt_idx.setter
    def mt_idx(self, value):
        """."""
        self['MTurnIdx-SP'] = int(value)

    @property
    def mt_idx_time(self):
        """."""
        return self['MTurnIdxTime-Mon']

    @property
    def mt_use_tbt_mask(self):
        """."""
        return bool(self['MTurnUseMask-Sts'])

    @mt_use_tbt_mask.setter
    def mt_use_tbt_mask(self, value):
        """."""
        self['MTurnUseMask-Sel'] = value

    @property
    def mt_tbt_mask_begin(self):
        """."""
        return self['MTurnMaskSplBeg-RB']

    @mt_tbt_mask_begin.setter
    def mt_tbt_mask_begin(self, value):
        """."""
        self['MTurnMaskSplBeg-SP'] = int(value)

    @property
    def mt_tbt_mask_end(self):
        """."""
        return self['MTurnMaskSplEnd-RB']

    @mt_tbt_mask_end.setter
    def mt_tbt_mask_end(self, value):
        """."""
        self['MTurnMaskSplEnd-SP'] = int(value)

    @property
    def trajx(self):
        """."""
        return self.mt_trajx_idx

    @property
    def trajy(self):
        """."""
        return self.mt_trajy_idx

    @property
    def sum(self):
        """."""
        return self.mt_sum_idx

    def cmd_mturn_acquire(self):
        """."""
        prop = 'MTurnAcquire-Cmd'
        val = self[prop]
        self[prop] = 1
        return self.wait(prop, val + 1)


class SISOFB(BOSOFB):
    """SOFB Device."""

    PROPERTIES_DEFAULT = BOSOFB.PROPERTIES_DEFAULT + (
        'SOFBMode-Sel',
        'SOFBMode-Sts',
        'KickRF-Mon',
        'DeltaKickRF-Mon',
        'DeltaKickRF-SP',
        'MaxDeltaKickRF-RB',
        'MaxDeltaKickRF-SP',
        'ManCorrGainRF-SP',
        'ManCorrGainRF-RB',
        'MeasRespMatKickRF-SP',
        'MeasRespMatKickRF-RB',
        'RFEnbl-Sel',
        'RFEnbl-Sts',
        'SlowOrbX-Mon',
        'SlowOrbY-Mon',
        'LoopState-Sts',
        'LoopState-Sel',
        'CorrSync-Sts',
        'CorrSync-Sel',
        'LoopPIDKpCH-SP',
        'LoopPIDKiCH-SP',
        'LoopPIDKdCH-SP',
        'LoopPIDKpCH-RB',
        'LoopPIDKiCH-RB',
        'LoopPIDKdCH-RB',
        'LoopPIDKpCV-SP',
        'LoopPIDKiCV-SP',
        'LoopPIDKdCV-SP',
        'LoopPIDKpCV-RB',
        'LoopPIDKiCV-RB',
        'LoopPIDKdCV-RB',
        'LoopPIDKpRF-SP',
        'LoopPIDKiRF-SP',
        'LoopPIDKdRF-SP',
        'LoopPIDKpRF-RB',
        'LoopPIDKiRF-RB',
        'LoopPIDKdRF-RB',
        'DriveFreqDivisor-SP',
        'DriveFreqDivisor-RB',
        'DriveFrequency-Mon',
        'DriveNrCycles-SP',
        'DriveNrCycles-RB',
        'DriveDuration-Mon',
        'DriveAmplitude-SP',
        'DriveAmplitude-RB',
        'DrivePhase-SP',
        'DrivePhase-RB',
        'DriveCorrIndex-SP',
        'DriveCorrIndex-RB',
        'DriveBPMIndex-SP',
        'DriveBPMIndex-RB',
        'DriveType-Sel',
        'DriveType-Sts',
        'DriveState-Sel',
        'DriveState-Sts',
        'DriveData-Mon',
    )

    @property
    def opmode(self):
        """."""
        return self['SOFBMode-Sts']

    @opmode.setter
    def opmode(self, value):
        self._enum_setter('SOFBMode-Sel', value, self._data.SOFBMode)

    @property
    def opmode_str(self):
        """."""
        return self._data.SOFBMode._fields[self['SOFBMode-Sts']]

    def cmd_change_opmode_to_multiturn(self, timeout=10):
        """."""
        mode = self._data.SOFBMode.MultiTurn
        self.opmode = mode
        ret = self.wait('SOFBMode-Sts', mode, timeout=timeout)
        if not ret:
            return False
        _time.sleep(1)  # Status PV updates at 2Hz
        return self.wait_orb_status_ok(timeout=timeout)

    def cmd_change_opmode_to_sloworb(self, timeout=10):
        """."""
        mode = self._data.SOFBMode.SlowOrb
        self.opmode = mode
        ret = self.wait('SOFBMode-Sts', mode, timeout=timeout)
        if not ret:
            return False
        _time.sleep(0.6)  # Status PV updates at 2Hz
        return self.wait_orb_status_ok(timeout=timeout)

    @property
    def loop_pid_ch_kp(self):
        """Loop PID Kp parameter for CH."""
        return self['LoopPIDKpCH-RB']

    @loop_pid_ch_kp.setter
    def loop_pid_ch_kp(self, value):
        self['LoopPIDKpCH-SP'] = value

    @property
    def loop_pid_ch_ki(self):
        """Loop PID Ki parameter for CH."""
        return self['LoopPIDKiCH-RB']

    @loop_pid_ch_ki.setter
    def loop_pid_ch_ki(self, value):
        self['LoopPIDKiCH-SP'] = value

    @property
    def loop_pid_ch_kd(self):
        """Loop PID Kd parameter for CH."""
        return self['LoopPIDKdCH-RB']

    @loop_pid_ch_kd.setter
    def loop_pid_ch_kd(self, value):
        self['LoopPIDKdCH-SP'] = value

    @property
    def loop_pid_cv_kp(self):
        """Loop PID Kp parameter for CV."""
        return self['LoopPIDKpCV-RB']

    @loop_pid_cv_kp.setter
    def loop_pid_cv_kp(self, value):
        self['LoopPIDKpCV-SP'] = value

    @property
    def loop_pid_cv_ki(self):
        """Loop PID Ki parameter for CV."""
        return self['LoopPIDKiCV-RB']

    @loop_pid_cv_ki.setter
    def loop_pid_cv_ki(self, value):
        self['LoopPIDKiCV-SP'] = value

    @property
    def loop_pid_cv_kd(self):
        """Loop PID Kd parameter for CV."""
        return self['LoopPIDKdCV-RB']

    @loop_pid_cv_kd.setter
    def loop_pid_cv_kd(self, value):
        self['LoopPIDKdCV-SP'] = value

    @property
    def loop_pid_rf_kp(self):
        """Loop PID Kp parameter for RF."""
        return self['LoopPIDKpRF-RB']

    @loop_pid_rf_kp.setter
    def loop_pid_rf_kp(self, value):
        self['LoopPIDKpRF-SP'] = value

    @property
    def loop_pid_rf_ki(self):
        """Loop PID Ki parameter for RF."""
        return self['LoopPIDKiRF-RB']

    @loop_pid_rf_ki.setter
    def loop_pid_rf_ki(self, value):
        self['LoopPIDKiRF-SP'] = value

    @property
    def loop_pid_rf_kd(self):
        """Loop PID Kd parameter for RF."""
        return self['LoopPIDKdRF-RB']

    @loop_pid_rf_kd.setter
    def loop_pid_rf_kd(self, value):
        self['LoopPIDKdRF-SP'] = value

    @property
    def drivests(self):
        """."""
        return self['DriveState-Sts']

    @property
    def drivetype(self):
        """."""
        return self['DriveType-Sts']

    @drivetype.setter
    def drivetype(self, value):
        self._enum_setter('DriveType-Sel', value, self._data.DriveType)

    @property
    def drivetype_str(self):
        """."""
        return self._data.DriveType._fields[self['DriveType-Sts']]

    @property
    def drivefreqdivisor(self):
        """."""
        return self['DriveFreqDivisor-RB']

    @drivefreqdivisor.setter
    def drivefreqdivisor(self, value):
        """."""
        self['DriveFreqDivisor-SP'] = value

    @property
    def drivefrequency_mon(self):
        """."""
        return self['DriveFrequency-Mon']

    @property
    def drivenrcycles(self):
        """."""
        return self['DriveNrCycles-RB']

    @drivenrcycles.setter
    def drivenrcycles(self, value):
        """."""
        self['DriveNrCycles-SP'] = value

    @property
    def driveduration_mon(self):
        """."""
        return self['DriveDuration-Mon']

    @property
    def driveamplitude(self):
        """."""
        return self['DriveAmplitude-RB']

    @driveamplitude.setter
    def driveamplitude(self, value):
        """."""
        self['DriveAmplitude-SP'] = value

    @property
    def drivephase(self):
        """."""
        return self['DrivePhase-RB']

    @drivephase.setter
    def drivephase(self, value):
        """."""
        self['DrivePhase-SP'] = value

    @property
    def drivecorridx(self):
        """."""
        return self['DriveCorrIndex-RB']

    @drivecorridx.setter
    def drivecorridx(self, value):
        """."""
        self['DriveCorrIndex-SP'] = value

    @property
    def drivebpmidx(self):
        """."""
        return self['DriveBPMIndex-RB']

    @drivebpmidx.setter
    def drivebpmidx(self, value):
        """."""
        self['DriveBPMIndex-SP'] = value

    @property
    def drivedata_time(self):
        """."""
        return self['DriveData-Mon'][::3]

    @property
    def drivedata_corr(self):
        """."""
        return self['DriveData-Mon'][1::3]

    @property
    def drivedata_bpm(self):
        """."""
        return self['DriveData-Mon'][2::3]

    @property
    def orbx(self):
        """."""
        return self['SlowOrbX-Mon'] if self._data.isring else None

    @property
    def orby(self):
        """."""
        return self['SlowOrbY-Mon'] if self._data.isring else None

    @property
    def kickrf(self):
        """."""
        return self['KickRF-Mon']

    @property
    def deltakickrf(self):
        """."""
        return self['DeltaKickRF-Mon']

    @deltakickrf.setter
    def deltakickrf(self, value):
        """."""
        self['DeltaKickRF-SP'] = value

    @property
    def maxdeltakickrf(self):
        """."""
        return self['MaxDeltaKickRF-RB']

    @maxdeltakickrf.setter
    def maxdeltakickrf(self, value):
        """."""
        self['MaxDeltaKickRF-SP'] = value

    @property
    def mancorrgainrf(self):
        """."""
        return self['ManCorrGainRF-RB']

    @mancorrgainrf.setter
    def mancorrgainrf(self, value):
        """."""
        self['ManCorrGainRF-SP'] = value

    @property
    def rfenbl(self):
        """."""
        if self._data.acc_idx == self._data.Rings.SI:
            return bool(self['RFEnbl-Sts'])

    @rfenbl.setter
    def rfenbl(self, value):
        """."""
        if self._data.acc_idx == self._data.Rings.SI:
            self['RFEnbl-Sel'] = value

    @property
    def measrespmat_kickrf(self):
        """."""
        return self['MeasRespMatKickRF-RB']

    @measrespmat_kickrf.setter
    def measrespmat_kickrf(self, value):
        self['MeasRespMatKickRF-SP'] = value

    @property
    def autocorrsts(self):
        """."""
        return self['LoopState-Sts']

    def cmd_turn_on_autocorr(self, timeout=None):
        """."""
        timeout = timeout or self._default_timeout
        if self.autocorrsts == self._data.LoopState.Closed:
            return True
        self['LoopState-Sel'] = self._data.LoopState.Closed
        return self.wait(
            'LoopState-Sts', self._data.LoopState.Closed, timeout=timeout
        )

    def cmd_turn_off_autocorr(self, timeout=None):
        """."""
        timeout = timeout or self._default_timeout
        if self.autocorrsts == self._data.LoopState.Open:
            return True
        self['LoopState-Sel'] = self._data.LoopState.Open
        return self.wait(
            'LoopState-Sts', self._data.LoopState.Open, timeout=timeout
        )

    @property
    def synckicksts(self):
        """Correction syncronization status."""
        return self['CorrSync-Sts']

    @synckicksts.setter
    def synckicksts(self, value):
        self._enum_setter('CorrSync-Sel', value, self._data.CorrSync)

    @property
    def synckicksts_str(self):
        """Correction syncronization status enum string."""
        return self._data.CorrSync._fields[self['CorrSync-Sts']]

    def cmd_turn_off_synckick(self, timeout=None):
        """Turn off correction synchronization."""
        timeout = timeout or self._default_timeout
        if self.synckicksts == self._data.CorrSync.Off:
            return True
        self['CorrSync-Sel'] = self._data.CorrSync.Off
        return self.wait(
            'CorrSync-Sts', self._data.CorrSync.Off, timeout=timeout
        )

    def cmd_turn_on_drive(self, timeout=None):
        """."""
        timeout = timeout or self._default_timeout
        if self.drivests == self._data.DriveState.Closed:
            return True
        self['DriveState-Sel'] = self._data.DriveState.Closed
        return self.wait(
            'DriveState-Sts', self._data.DriveState.Closed, timeout=timeout
        )

    def cmd_turn_off_drive(self, timeout=None):
        """."""
        timeout = timeout or self._default_timeout
        if self.drivests == self._data.DriveState.Open:
            return True
        self['DriveState-Sel'] = self._data.DriveState.Open
        return self.wait_drive(timeout=timeout)

    def wait_drive(self, timeout=None):
        """."""
        timeout = timeout or self._default_timeout_respm
        return self.wait(
            'DriveState-Sts', self._data.DriveState.Open, timeout=timeout
        )

    @staticmethod
    def si_calculate_bumps(
        orbx,
        orby,
        subsec,
        agx=0,
        agy=0,
        psx=0,
        psy=0,
        n_bpms_out=3,
        minsingval=0.2,
    ):
        """."""
        return _si_calculate_bump(
            orbx,
            orby,
            subsec,
            agx=agx,
            agy=agy,
            psx=psx,
            psy=psy,
            n_bpms_out=n_bpms_out,
            minsingval=minsingval,
        )

    @property
    def trajx(self):
        """."""
        if self.opmode == self._data.SOFBMode.MultiTurn:
            return self.mt_trajx_idx
        return self.sp_trajx

    @property
    def trajy(self):
        """."""
        if self.opmode == self._data.SOFBMode.MultiTurn:
            return self.mt_trajy_idx
        return self.sp_trajy

    @property
    def sum(self):
        """."""
        if self.opmode == self._data.SOFBMode.MultiTurn:
            return self.mt_sum_idx
        return self.sp_sum



"""Tune devices."""

from .device import Device as _Device, DeviceSet as _DeviceSet


class TuneFrac(_Device):
    """Tune Frac device."""

    DEF_TIMEOUT = 1  # [s]

    class DEVICES:
        """Devices names."""

        SI_H = 'SI-Glob:DI-Tune-H'
        SI_V = 'SI-Glob:DI-Tune-V'
        ALL = (SI_H, SI_V)

    PROPERTIES_DEFAULT = (
        'TuneFrac-Mon',
        'Enbl-Sel',
        'Enbl-Sts',
        'Span-RB',
        'Span-SP',
        'RevN-RB',
        'RevN-SP',
        'FreqOff-RB',
        'FreqOff-SP',
    )

    def __init__(self, devname, props2init='all'):
        """Init."""
        if devname not in TuneFrac.DEVICES.ALL:
            raise NotImplementedError(devname)
        super().__init__(devname, props2init=props2init)

    @property
    def tune(self):
        """Tune Frac."""
        return self['TuneFrac-Mon']

    @property
    def span(self):
        """Span [kHz]."""
        return self['Span-RB']

    @span.setter
    def span(self, value):
        self['Span-SP'] = value

    @property
    def rev_harmonic(self):
        """Revolution harmonic."""
        return self['RevN-RB']

    @rev_harmonic.setter
    def rev_harmonic(self, value):
        """Revolution harmonic."""
        if not 0 <= int(value) <= 864:
            raise ValueError('rev_harmonic must be in range [0, 864]')
        self['RevN-SP'] = int(value)

    @property
    def center_frequency(self):
        """Center frequency offset [kHz]."""
        return self['FreqOff-RB']

    @center_frequency.setter
    def center_frequency(self, value):
        self['FreqOff-SP'] = value

    @property
    def enable(self):
        """Enable status."""
        return self['Enbl-Sts']

    def cmd_enable(self, timeout=DEF_TIMEOUT):
        """Enable."""
        self['Enbl-Sel'] = 1
        return self.wait('Enbl-Sts', value=1, timeout=timeout)

    def cmd_disable(self, timeout=DEF_TIMEOUT):
        """Disable."""
        self['Enbl-Sel'] = 0
        return self.wait('Enbl-Sts', value=0, timeout=timeout)


class TuneProc(_Device):
    """Tune Proc device."""

    class DEVICES:
        """Devices names."""

        SI_H = 'SI-Glob:DI-TuneProc-H'
        SI_V = 'SI-Glob:DI-TuneProc-V'
        ALL = (SI_H, SI_V)

    PROPERTIES_DEFAULT = ('Trace-Mon',)

    def __init__(self, devname, props2init='all'):
        """Init."""
        # check if device exists
        if devname not in TuneProc.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, props2init=props2init)

    @property
    def tune_wfm(self):
        """Tune waveform."""
        return self['Trace-Mon']


class Tune(_DeviceSet):
    """Tune device."""

    class DEVICES:
        """Devices names."""

        SI = 'SI-Glob:DI-Tune'
        ALL = (SI,)

    def __init__(self, devname=None, props2init='all'):
        """Init."""
        # check if device exists
        if devname is None:
            devname = Tune.DEVICES.SI
        if devname not in Tune.DEVICES.ALL:
            raise NotImplementedError(devname)

        isall = isinstance(props2init, str) and props2init.lower() == 'all'
        if not isall and props2init:
            raise ValueError(
                "props2init must be 'all' or bool(props2init) == False"
            )

        tune_frac_h = TuneFrac(TuneFrac.DEVICES.SI_H, props2init=props2init)
        tune_frac_v = TuneFrac(TuneFrac.DEVICES.SI_V, props2init=props2init)
        tune_proc_h = TuneProc(TuneProc.DEVICES.SI_H, props2init=props2init)
        tune_proc_v = TuneProc(TuneProc.DEVICES.SI_V, props2init=props2init)

        devices = (tune_frac_h, tune_frac_v, tune_proc_h, tune_proc_v)

        # call base class constructor
        super().__init__(devices, devname=devname)

    @property
    def tunex(self):
        """Tune Frac X."""
        return self.devices[0].tune

    @property
    def tuney(self):
        """Tune Frac Y."""
        return self.devices[1].tune

    @property
    def spanx(self):
        """Span X [kHz]."""
        return self.devices[0].span

    @spanx.setter
    def spanx(self, value):
        self.devices[0].span = value

    @property
    def spany(self):
        """Span Y [kHz]."""
        return self.devices[1].span

    @spany.setter
    def spany(self, value):
        self.devices[1].span = value

    @property
    def rev_harmonicx(self):
        """Revolution harmonic X."""
        return self.devices[0].rev_harmonic

    @rev_harmonicx.setter
    def rev_harmonicx(self, value):
        self.devices[0].rev_harmonic = value

    @property
    def rev_harmonicy(self):
        """Revolution harmonic Y."""
        return self.devices[1].rev_harmonic

    @rev_harmonicy.setter
    def rev_harmonicy(self, value):
        """Revolution harmonic Y."""
        self.devices[1].rev_harmonic = value

    @property
    def center_frequencyx(self):
        """Center frequency X."""
        return self.devices[0].center_frequency

    @center_frequencyx.setter
    def center_frequencyx(self, value):
        self.devices[0].center_frequency = value

    @property
    def center_frequencyy(self):
        """Center frequency Y."""
        return self.devices[1].center_frequency

    @center_frequencyy.setter
    def center_frequencyy(self, value):
        self.devices[1].center_frequency = value

    @property
    def tunex_wfm(self):
        """Tune waveform X."""
        return self.devices[2].tune_wfm

    @property
    def tuney_wfm(self):
        """Tune waveform Y."""
        return self.devices[3].tune_wfm

    @property
    def enablex(self):
        """Tune X enable status."""
        return self.devices[0].enable

    @property
    def enabley(self):
        """Tune Y enable status."""
        return self.devices[1].enable

    def cmd_enablex(self, timeout=TuneFrac.DEF_TIMEOUT):
        """Enable tune X."""
        return self.devices[0].cmd_enable(timeout=timeout)

    def cmd_enabley(self, timeout=TuneFrac.DEF_TIMEOUT):
        """Enable tune Y."""
        return self.devices[1].cmd_enable(timeout=timeout)

    def cmd_disablex(self, timeout=TuneFrac.DEF_TIMEOUT):
        """Disable tune X."""
        return self.devices[0].cmd_disable(timeout=timeout)

    def cmd_disabley(self, timeout=TuneFrac.DEF_TIMEOUT):
        """Disable tune Y."""
        return self.devices[1].cmd_disable(timeout=timeout)


class TuneCorr(_Device):
    """TuneCorr device."""

    class DEVICES:
        """Devices names."""

        SI = 'SI-Glob:AP-TuneCorr'
        ALL = (SI,)

    PROPERTIES_DEFAULT = (
        'DeltaTuneX-SP',
        'DeltaTuneX-RB',
        'DeltaTuneY-SP',
        'DeltaTuneY-RB',
        'SetNewRefKL-Cmd',
        'ApplyDelta-Cmd',
    )

    def __init__(self, devname=None, props2init='all'):
        """Init."""
        if devname is None:
            devname = TuneCorr.DEVICES.SI
        if devname not in TuneCorr.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, props2init=props2init)

    @property
    def delta_tunex(self):
        """Delta Tune X."""
        return self['DeltaTuneX-RB']

    @delta_tunex.setter
    def delta_tunex(self, value):
        self['DeltaTuneX-SP'] = value

    @property
    def delta_tuney(self):
        """Delta Tune Y."""
        return self['DeltaTuneY-RB']

    @delta_tuney.setter
    def delta_tuney(self, value):
        self['DeltaTuneY-SP'] = value

    def cmd_update_reference(self):
        """Update reference tunes."""
        self['SetNewRefKL-Cmd'] = 1
        return True

    def cmd_apply_delta(self):
        """Apply delta tunes."""
        self['ApplyDelta-Cmd'] = 1
        return True



"""."""

import numpy as _np

from ..currinfo.csdev import Const as _Const
from .device import Device as _Device


class FPMOsc(_Device):
    """."""

    OffOn = _Const.OffOn

    class DEVICES:
        """Devices names."""

        SI = 'SI-Glob:DI-FPMOsc'
        ALL = (SI, )

    PROPERTIES_DEFAULT = (
        'UpdateTime-SP',
        'UpdateTime-RB',
        'FiducialOffset-SP',
        'FiducialOffset-RB',
        'FillPatternRef-SP',
        'FillPatternRef-RB',
        'UniFillEqCurrent-Mon',
        'ErrorRelStd-Mon',
        'ErrorKLDiv-Mon',
        'FillPatternRef-Mon',
        'FillPattern-Mon',
        # 'TimeOffset-Mon',
        # 'Time-Mon',
        # 'Raw-Mon',
        # 'RawAmp-Mon',
        # 'RawTime-Mon',
    )

    def __init__(self, devname=None, props2init='all'):
        """."""
        # check if device exists
        devname = devname or FPMOsc.DEVICES.SI
        if devname not in FPMOsc.DEVICES.ALL:
            raise NotImplementedError(devname)
        super().__init__(devname, props2init=props2init)

    @property
    def update_time(self):
        """."""
        return self['UpdateTime-RB']

    @update_time.setter
    def update_time(self, value):
        """."""
        self['UpdateTime-SP'] = value

    @property
    def fiducial_offset(self):
        """."""
        return self['FiducialOffset-RB']

    @fiducial_offset.setter
    def fiducial_offset(self, value):
        self['FiducialOffset-SP'] = value

    @property
    def fill_pattern_ref(self):
        """."""
        return self['FillPatternRef-RB']

    @fill_pattern_ref.setter
    def fill_pattern_ref(self, value):
        """."""
        if not isinstance(value, (_np.ndarray, list, tuple)):
            raise TypeError('Value must be list, tuple or numpy.ndarray.')
        value = _np.array(value)
        if value.size != _Const.FP_HARM_NUM:
            raise ValueError('Value size must be the harmonic number.')
        self['FillPatternRef-SP'] = value

    @property
    def fill_pattern_ref_mon(self):
        """."""
        return self['FillPatternRef-Mon']

    @property
    def fill_pattern(self):
        """."""
        return self['FillPattern-Mon']

    @property
    def uniform_fill_equiv_current(self):
        """."""
        return self['UniFillEqCurrent-Mon']

    @property
    def error_relative_std(self):
        """."""
        return self['ErrorRelStd-Mon']

    @property
    def error_kl_divergence(self):
        """."""
        return self['ErrorKLDiv-Mon']

    @property
    def time_offset(self):
        """."""
        return self['TimeOffset-Mon']

    @property
    def fill_pattern_time(self):
        """."""
        return self['Time-Mon']

    @property
    def fill_pattern_raw(self):
        """."""
        return self['Raw-Mon']

    @property
    def fill_pattern_raw_time(self):
        """."""
        return self['RawTime-Mon']
